

         /*********************************************************/
         /*                                                       */
         /*                 COPYRIGHT (C) 1988 BY                 */
         /* CAD Group of Dipartimento di Automatica e Informatica */
         /*             Politecnico di Torino, Italy.             */
         /*                  ALL RIGHTS RESERVED.                 */
         /*                                                       */
         /* This software is furnished under a license and may be */
         /* used and copied  only in accordance with the terms of */
         /* such license. Unauthorized copies will be persecuted. */
         /*                                                       */
         /*********************************************************/





/******************************************************************************

    Filename:         tab2ws.c

    Author:           Pier Luca Montessoro

    Creation date:    16 January 1989

    Contents:         WSL file generator from tabular patterns


                          *** DOCUMENTATION ***


 Generated source file format:

Header:

   $input
   out1
   out2
   ...

   $output
   out1
   out2
   ...

   $clock

   $begin

Statements

   time       $strobeperiod  time

   time       input  value

   time       $end

******************************************************************************/



#include "env.h"
#include "common.h"
#include "util.h"
#include "loadcode.h"
#include "topology.h"



#define  LINKER_VERSION       "3.0"




/* topological variables */


/* delay table element structure definition */

typedef struct Delay_Table_Element
        {
           unsigned long  rise;
           unsigned long  fall;
        }  DELAY_TABLE_ELEMENT;


DELAY_TABLE_ELEMENT    *delay_table;
TOPOLOGICAL_ELEMENT    *topology_vector;
FANIN_ELEMENT          *fanin_vector;
FANOUT_ELEMENT         *fanout_vector;

unsigned long          num_of_elements;
unsigned long          max_level_number;
unsigned long          max_eval_link_number;




#define  MAX_INP_NAME_LENGTH   16
#define  MAX_PERIOD_LENGTH     64

#define  TAB_EXTENSION      "dgp"
#define  WSL_EXTENSION      "dws"
#define  SYM_TAB_EXTENSION   "smt"



/* symbol table element structure definition */

typedef struct Sym_Tab
        {
           unsigned char   is_output;
           unsigned char   is_input;
           char            last_value;
           char            unused;
           unsigned long   last_time;
           unsigned long   index;
           unsigned long   type;
           unsigned long   bit;
           char            name [MAX_NAME_LENGTH];

        }  SYM_TAB;


/* symbol table */

SYM_TAB                *st_address;
unsigned long          sym_tab_dim;


/* network file */

FILE                   *net_file;


/* output WSL file */

FILE                   *wsl_file;


/* input tabular file */

FILE                   *tab_file;



/******************************************************************************

    Function name:    main

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          main

 ******************************************************************************/


main (argc, argv)

int   argc;
char  *argv[];

{

   /* called function */

   unsigned long         strtolong();


   /* local variables */

   char                  net_file_name [MAX_FILE_NAME_LENGTH];
   char                  wsl_file_name [MAX_FILE_NAME_LENGTH];
   char                  tab_file_name [MAX_FILE_NAME_LENGTH];
   unsigned long         quiescence_time;


   /* beginning of executable part */

   if (argc != 4)
   {
      printf ("Wrong number of arguments\n");
      printf 
      ("(1st: net filename, 2nd: pattern filename, 3rd: quiescence time).\n");
      good_bye (FALSE, TRUE);
   }


   /* open the net, source and the output files */

   strcpy (net_file_name, argv[1]);
   strcat (net_file_name, ".");
   strcat (net_file_name, NET_EXTENSION);

   strcpy (wsl_file_name, argv[1]);
   strcat (wsl_file_name, ".");
   strcat (wsl_file_name, WSL_EXTENSION);

   strcpy (tab_file_name, argv[1]);
   strcat (tab_file_name, ".");
   strcat (tab_file_name, TAB_EXTENSION);

   if ((wsl_file = fopen (wsl_file_name, "w")) == NULL)
   {
      printf ("I cannot open the output file.\n");
      good_bye (TRUE, TRUE);
   }

   if ((tab_file = fopen (tab_file_name, "r")) == NULL)
   {
      printf ("I cannot open the input pattern file.\n");
      good_bye (TRUE, TRUE);
   }

   if ((net_file = fopen (net_file_name, "r")) == NULL)
   {
      printf ("I cannot open the network file.\n");
      good_bye (FALSE, TRUE);
   }

   quiescence_time = strtolong (argv[3]);


   load_net ();
   load_sym_tab (argv[1]);

   pat_gen (quiescence_time);


   printf ("Output file %s written.\n", wsl_file_name);

   good_bye (TRUE, FALSE);

}  /* end of main */



/******************************************************************************

    Function name:    pat_gen

    Authors:          Pier Luca Montessoro

    Creation date:    16 January 1989

    Purpose:          to generate the wsl file

 ******************************************************************************/


#define MAX_INP_NUM       256


pat_gen (quiescence_time)

unsigned long  quiescence_time;

{

   /* called function */

   char           *get_memory();
   unsigned long  el_index();
   unsigned long  st_index();
   char           *el_name();
   unsigned char  is_integer();
   unsigned char  is_value();
   char           *uppercase();


   /* local variables */

   char           buffer [BUFFER_SIZE];
   char           tmp_name [MAX_NAME_LENGTH];
   unsigned char  ck_present, res_present, res_active;
   unsigned long  i;
   unsigned long  cnt;
   unsigned long  inp_num;
   unsigned long  tmp_inp_num;
   unsigned long  time;
   char           inp_list [MAX_INP_NUM] [MAX_INP_NAME_LENGTH];


   /* beginning of executable part */


   /* collect the inputs */

   inp_num = 0;
   for (i = 0; i < num_of_elements; i++)
   {
      if (topology_vector[i].position == INPUT)
      { 
         strcpy (tmp_name, el_name (i));
         if (inp_num >= MAX_INP_NUM)
         {
            printf ("Too many inputs.\n");
            good_bye (TRUE, TRUE);
         }
         /*checking for clocks --Jamie*/
	 fprintf(stderr,"tmp_name = %s\n",tmp_name);
         if (strcmp(tmp_name,"ck")){ 
	    fprintf(stderr,"inside if\n");
            strcpy (inp_list[inp_num++], tmp_name);
	 }
      }
   }


   /* write header */
   fprintf (wsl_file, "\n$input\n");

   for (i = 0; i < inp_num; i++)
   {
      fprintf (wsl_file, "   %s\n", inp_list[i]);
   }


   fprintf (wsl_file, "\n$output\n");
   for (i = 0; i < num_of_elements; i++)
   {
      if (topology_vector[i].detection_point)
      {
         fprintf (wsl_file, "   %s\n", el_name(i));
      }
   }

   fprintf (wsl_file, "\n$clock\n");


   fprintf (wsl_file, "\n$begin\n");


   fprintf (wsl_file, "\n0 $strobeperiod %d\n\n", quiescence_time);


   /* write patterns */

   time = 0;
   while (fscanf (tab_file, "%s", buffer) != EOF)
   {
      if ((tmp_inp_num = strlen (buffer)) > inp_num)
      {
         printf ("Too many inputs in pattern file.\n");
         good_bye (TRUE, TRUE);
      }

     /* if (tmp_inp_num < (inp_num-1))
      {
         printf ("Too few inputs in pattern file.\n");
         good_bye (TRUE, TRUE);
      }
     */

      for (i = 0; i < inp_num; i++)
      {
         fprintf (wsl_file, "%lu %s %c\n", 
                  time, inp_list[i], buffer[i]);
      }


      time += quiescence_time;

   }


   time += quiescence_time;


   fprintf (wsl_file, "%lu $end\n\n", time);


   printf ("End time: %lu.\n", time);

   return;

}  /* end of pat_gen */



/******************************************************************************

    Function name:    good_bye

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          exit function

 ******************************************************************************/


good_bye (open_files, error_found)

unsigned char  open_files;
unsigned char*  error_found;

{

   /* beginning of executable part */

   if (open_files)
   {
      fclose (net_file);
      fclose (wsl_file);
      fclose (tab_file);
   }

   if (error_found)
   {
      printf ("I am sorry, but I cannot proceed. Try again.\n");
      
      printf (error_found);
   }
   else
   {
      printf ("Job terminated. Good-bye!");
   }

   exit (1);

}  /* end of good_bye */



/******************************************************************************

    Function name:    get_memory

    Authors:          Pier Luca Montessoro

    Creation date:    25 January 1988

    Purpose:          memory allocation for network data structures

 ******************************************************************************/


char  *get_memory (number, size)

unsigned long  number;
unsigned long  size;

{

   /* local variable */

   char   *alloc_pointer;


   /* beginning of executable part */

   alloc_pointer = malloc (number * size);


   if (alloc_pointer == NULL)
   {
      printf ("No memory is available for topological data allocation.\n");
      good_bye (TRUE, TRUE);
   }

   return (alloc_pointer);

}  /* end of get_memory */



/******************************************************************************

    Function name:    uppercase

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to convert a string to uppercase

 ******************************************************************************/


char *uppercase (name)

char  *name;

{

   /* local variable */

   unsigned long  i;


   /* beginning of executable part */

   for (i = 0; name[i] != '\0'; name [i] = toupper (name [i]), i++);
   return (name);

}  /* end of uppercase */



/******************************************************************************

    Function name:    load_sym_tab

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to load the symbol table of the current experiment

 ******************************************************************************/


load_sym_tab (exp_name)

char  *exp_name;

{

   /* local variables */

   char                  file_name [MAX_FILE_NAME_LENGTH];
   FILE                  *sym_tab_file;
   char                  *st_name;
   unsigned long         el_type;
   char                  dummy_str [MAX_NAME_LENGTH];
   unsigned long         dummy_num;
   unsigned long         mod_size;
   unsigned long         n_modules;
   unsigned long         it_size;
   unsigned long         t_index;
   unsigned long         bit;
   unsigned long         i;
   unsigned long         j;


   /* beginning of executable part */

   /* open the file containing the symbol table */

   sprintf (file_name, "%s.%s", exp_name, SYM_TAB_EXTENSION);
   
   if ((sym_tab_file = fopen (file_name, "r")) == NULL)
   {
      printf ("Symbol table file not found\n");
      good_bye(TRUE, TRUE);
   }


   /* read the number of modules */
   fscanf (sym_tab_file, "%ld %ld", &n_modules, &dummy_num);


   /* reach the main module (the first one) */

   for (j = 0; j < n_modules; j++)
   {

      /* read the size of the interface table */   
      fscanf (sym_tab_file, "%s %ld %ld %s %ld %ld %ld %ld", 
              dummy_str, &dummy_num, &dummy_num, dummy_str, 
              &dummy_num, &dummy_num, &dummy_num, 
              &it_size);

      /* skip the interface table (it should be empty for the main module) */
      for (i = 0; i < it_size; i++)
      {
         fscanf (sym_tab_file, "%ld %s", &dummy_num, dummy_str);
      }
   }


   /* read the module name and the size */
   fscanf (sym_tab_file, "%s %ld", dummy_str, &mod_size);

   /* alloc the memory for the symbol table */
   if ((st_address = (SYM_TAB *) 
      get_memory (mod_size, (unsigned long) sizeof (SYM_TAB))) == NULL)  
 
   {
       printf ("I cannot alloc memory for the symbol table.\n");
      fclose (sym_tab_file);
      good_bye (TRUE, TRUE);
   }

   sym_tab_dim = mod_size;

   /* load the names */
   for (i = 0; i < mod_size; i++)
   {
      fscanf (sym_tab_file, "%ld %ld %ld %s", &t_index, &bit, &el_type, 
	      dummy_str);
      st_address[i].index = t_index;
      st_address[i].is_output = FALSE;
      st_address[i].is_input = FALSE;
      st_address[i].last_value = 'X';
      st_address[i].last_time = MAX_UNSIGNED;
      st_address[i].bit = bit;
      st_address[i].type = el_type;

      /* convert the name to uppercase */
      st_name = st_address[i].name;
      for (j = 0; dummy_str[j] != '\0'; j++)
      {
         st_name[j] = tolower (dummy_str[j]);
      }
      st_name[j] = '\0';
   }

   fclose (sym_tab_file);

   printf ("Symbol table loaded\n");

   return;

}  /* end of load_sym_tab */



/******************************************************************************

    Function name:    el_index

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to find the index of an element

 ******************************************************************************/


unsigned long  el_index (name)

char  *name;

{

   /* local variables */

   unsigned long  i;


   /* beginning of executable part */

   for (i = 0; i < sym_tab_dim; i++)
   {
      if (!strcmp (st_address[i].name, name))
      {
         return (st_address[i].index);
      }
   }


   printf ("I cannot find element %s in the main symbol table.\n", name);

   good_bye (TRUE, TRUE);


}  /* end of el_index */



/******************************************************************************

    Function name:    st_index

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to find the index of an element

 ******************************************************************************/


unsigned long  st_index (name)

char  *name;

{

   /* local variables */

   unsigned long  i;


   /* beginning of executable part */

   for (i = 0; i < sym_tab_dim; i++)
   {
      if (!strcmp (st_address[i].name, name))
      {
         return (i);
      }
   }


   printf ("I cannot find element %s in the main symbol table.\n", name);

   good_bye (TRUE, TRUE);


}  /* end of st_index */



/******************************************************************************

    Function name:    el_name

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to return the name of an element (given its 
                      topological index)

 ******************************************************************************/


char *el_name (index)

unsigned long  index;

{

   /* local variables */

   unsigned long  i;
   unsigned long         tmp_index = MAX_UNSIGNED;


   /* beginning of executable part */

   for (i = 0; i < sym_tab_dim; i++)
   {
      if (index == st_address[i].index)
      {
         
         if (st_address[i].type != MZ_GATE)
         {
            return (st_address[i].name);
         }
         else
         {
            /* search for the wire */
            tmp_index = i;
         }
      }
   }


   /* not found as not GATE */

   if (tmp_index != MAX_UNSIGNED)
   {
      return (st_address[tmp_index].name);
   }


   /* not found */

   return ("");


}  /* end of el_name */



/******************************************************************************

    Function name:    is_value

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to check the type of a string

 ******************************************************************************/


unsigned char is_value (str)

char  *str;

{

   if ((str[1] != '\0') || 
       ((str[0] != '0') && (str[0] != '1') && 
        (str[0] != 'X') && (str[0] != 'Z')))
   {
      return (FALSE);
   }
   else
   {
      return (TRUE);
   }

}  /* end of is_value */



/******************************************************************************

    Function name:    strtolong

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to convert the timing string into a long int

 ******************************************************************************/


unsigned long  strtolong (str)

char  *str;

{

   /* local variables */

   unsigned long   i;
   unsigned long   len;
   unsigned long   res;


   /* beginning of executable part */

   if ((len = strlen (str)) == 0)
   {
      printf ("I cannot understand the timing information.\n");
      good_bye (TRUE, TRUE);
   }

   for (i = 0, res = 0; i < len; i++)
   {
      if ((str[i] < '0') || (str[i] > '9'))
      {
         printf ("I cannot understand the timing information.\n");
         good_bye (TRUE, TRUE);
      }

      res = (res * 10) + str[i] - '0';
   }

   if (res == 0)
   {
      printf ("Bad timing information.\n");
      good_bye (TRUE, TRUE);
   }


   return (res);

}  /* end of strtolong */



/******************************************************************************

    Function name:    load_net

    Authors:          Pier Luca Montessoro

    Creation date:    25 January 1988

    Purpose:          network data structure creation

 ******************************************************************************/


load_net ()

{

   /* called function */
   char                  *get_memory();

   /* local variables */
   char                  net_file_name [MAX_FILE_NAME_LENGTH];
   MACHINE               *ref_allocation_pointer;
   TOPOLOGICAL_ELEMENT   *second_pass = (TOPOLOGICAL_ELEMENT *) NULL;
   unsigned long          num_of_machines;


   /* delay table declaration */
   typedef struct Delay_Table_Element
   {
      unsigned long  rise;
      unsigned long  fall;
   }  DELAY_TABLE_ELEMENT;

   DELAY_TABLE_ELEMENT   *delay_table;


   /* general information variables */
   char              linker_version [MAX_NAME_LENGTH];
   char              circuit_name   [MAX_NAME_LENGTH];


   /* symbol table and module table offsets */
   long int          sym_tab_offset;
   long int          mod_tab_offset;

   /* number of warnings */
   unsigned long     warnings;

   /* information related to the topological elements */
   unsigned long     p_o_num;
   unsigned long     p_o_size;
   unsigned long     p_i_num;
   unsigned long     p_i_size;
   unsigned long     p_b_num;
   unsigned long     p_b_size;
   unsigned long     constants_num;
   unsigned long     nodes_num;
   unsigned long     wires_num;
   unsigned long     regi_num;
   unsigned long     wires_size;
   unsigned long     regi_size;
   unsigned long     transistors_num;
   unsigned long     gates_num;
   unsigned long     rt_assignement_num;
   unsigned long     rt_condition_num;
   unsigned long     memory_num;
   unsigned long     locvar_num;
   unsigned long     proc_num;
   unsigned long     unused;
   unsigned long     gates_array_num;

   /* delays */
   unsigned long     delays_num;
   unsigned long     delay_unit;
   unsigned long     min_delay;
   unsigned long     max_delay;

   /* initialization table */
   unsigned long     init_num;

   /* C-modules and functions */
   unsigned long     c_modules_num;
   unsigned long     c_functions_num;

   /* maximum parameters */
   unsigned long     max_strength;
   unsigned long     max_size;
   unsigned long     max_fanout;
   unsigned long     max_atpg_level;

   /* link date and time */
   unsigned long     link_date;
   unsigned long     link_time;
   unsigned long     c_link_date;

   /* MOZART internal information */
   long              vdd_offset;
   long              gnd_offset;
   unsigned long     static_size;

   /* flags */
   unsigned long     atpg_flag;
   unsigned long     sim_flag;
   unsigned long     gate_array_flag;
   unsigned long     case_sensitive_flag;

   /* topological data structure sizes */
   unsigned long     tot_elements_num;
   unsigned long     tot_input_num;
   unsigned long     tot_fanin_num;
   unsigned long     tot_output_num;
   unsigned long     tot_fanout_num;
   unsigned long     tot_bidir_num;

   /* averages values */
   unsigned long     nodes_fo;
   unsigned long     nodes_fi;
   unsigned long     nodes_conn;
   unsigned long     wires_fo;
   unsigned long     wires_fi;
   unsigned long     regi_fo;
   unsigned long     regi_fi;
   unsigned long     gates_fo;
   unsigned long     gates_fi;
   unsigned long     rt_a_fo;
   unsigned long     rt_a_fi;
   unsigned long     rt_c_fo;
   unsigned long     rt_c_fi;


   /* beginning of executable part */


  /* read in the description file (*.dee) */

   fscanf (net_file, "%s", linker_version);	
   if (strcmp (linker_version, LINKER_VERSION))
   {
      fclose (net_file);
      good_bye (FILE_ERROR, "the linker version is not compatible with me");
   }

   fscanf (net_file, "%s", circuit_name); 
   fscanf (net_file, "%ld", &sym_tab_offset);
   fscanf (net_file, "%ld", &mod_tab_offset);
   fscanf (net_file, "%ld", &warnings);	
   fscanf (net_file, "%ld", &p_i_num); 
   fscanf (net_file, "%ld", &p_i_size);
   fscanf (net_file, "%ld", &p_o_num); 
   fscanf (net_file, "%ld", &p_o_size);
   fscanf (net_file, "%ld", &p_b_num);
   fscanf (net_file, "%ld", &p_b_size);
   fscanf (net_file, "%ld", &constants_num);
   fscanf (net_file, "%ld", &nodes_num); 
   fscanf (net_file, "%ld", &wires_num); 
   fscanf (net_file, "%ld", &wires_size);
   fscanf (net_file, "%ld", &regi_num);	
   fscanf (net_file, "%ld", &regi_size);
   fscanf (net_file, "%ld", &transistors_num);
   fscanf (net_file, "%ld", &gates_num);
   fscanf (net_file, "%ld", &rt_assignement_num);
   fscanf (net_file, "%ld", &rt_condition_num);
   fscanf (net_file, "%ld", &memory_num);
   fscanf (net_file, "%ld", &locvar_num);
   fscanf (net_file, "%ld", &proc_num);	
   fscanf (net_file, "%ld", &unused);			
   fscanf (net_file, "%ld", &gates_array_num);
   fscanf (net_file, "%ld", &delays_num);
   fscanf (net_file, "%ld", &delay_unit);
   fscanf (net_file, "%ld", &min_delay);
   fscanf (net_file, "%ld", &max_delay);
   fscanf (net_file, "%ld", &init_num);
   fscanf (net_file, "%ld", &unused);			
   fscanf (net_file, "%ld", &unused);
   fscanf (net_file, "%ld", &c_modules_num);
   fscanf (net_file, "%ld", &c_functions_num);
   fscanf (net_file, "%ld", &max_strength);
   fscanf (net_file, "%ld", &max_size);	
   fscanf (net_file, "%ld", &max_fanout);
   fscanf (net_file, "%ld", &max_atpg_level);
   fscanf (net_file, "%ld", &vdd_offset);
   fscanf (net_file, "%ld", &gnd_offset);
   fscanf (net_file, "%ld", &link_date);
   fscanf (net_file, "%ld", &link_time);
   fscanf (net_file, "%ld", &c_link_date);
   fscanf (net_file, "%ld", &static_size);
   fscanf (net_file, "%ld", &atpg_flag);
   fscanf (net_file, "%ld", &sim_flag);
   fscanf (net_file, "%ld", &gate_array_flag);
   fscanf (net_file, "%ld", &case_sensitive_flag);
   fscanf (net_file, "%ld", &tot_elements_num);	
   fscanf (net_file, "%ld", &tot_fanin_num);
   fscanf (net_file, "%ld", &tot_input_num);
   fscanf (net_file, "%ld", &tot_fanout_num);
   fscanf (net_file, "%ld", &tot_output_num);
   fscanf (net_file, "%ld", &tot_bidir_num);
   fscanf (net_file, "%ld", &nodes_fo);
   fscanf (net_file, "%ld", &nodes_fi);
   fscanf (net_file, "%ld", &nodes_conn);
   fscanf (net_file, "%ld", &wires_fo);
   fscanf (net_file, "%ld", &wires_fi);
   fscanf (net_file, "%ld", &regi_fo);
   fscanf (net_file, "%ld", &regi_fi);
   fscanf (net_file, "%ld", &gates_fo);
   fscanf (net_file, "%ld", &gates_fi);
   fscanf (net_file, "%ld", &rt_a_fo);
   fscanf (net_file, "%ld", &rt_a_fi);
   fscanf (net_file, "%ld", &rt_c_fo);
   fscanf (net_file, "%ld", &rt_c_fi);



#if SHOW_HEADER   /*Display the header section of the *.dee file*/
   {
      /* external function */
      extern int  out_message();
      extern int  sprintf();

      /* local variable */
      char  header_info [MAX_MESSAGE_LENGTH];

      /* beginning of executable part */
      sprintf (header_info, "linker_version : %s\n",  linker_version);
      out_message (header_info);
      sprintf (header_info, "circuit_name : %6s\n", circuit_name);
      out_message (header_info);
      sprintf (header_info, "warnings : %6lu\n", warnings);
      out_message (header_info);
      sprintf (header_info, "primary_output_num : %6lu\n", p_o_num);
      out_message (header_info);
      sprintf (header_info, "primary_output_size : %6lu\n", p_o_size);
      out_message (header_info);
      sprintf (header_info, "primary_input_num : %6lu\n", p_i_num);
      out_message (header_info);
      sprintf (header_info, "primary_input_size : %6lu\n", p_i_size);
      out_message (header_info);
      sprintf (header_info, "primary_bidirectional_num : %6lu\n", p_b_num);
      out_message (header_info);
      sprintf (header_info, "primary_bidirectional_size : %6lu\n", p_b_size);
      out_message (header_info);
      sprintf (header_info, "constants_num : %6lu\n", constants_num);
      out_message (header_info);
      sprintf (header_info, "nodes_num : %6lu\n", nodes_num);
      out_message (header_info);
      sprintf (header_info, "wires_num : %6lu\n", wires_num);
      out_message (header_info);
      sprintf (header_info, "wires_size : %6lu\n", wires_size);
      out_message (header_info);
      sprintf (header_info, "regi_num : %6lu\n", regi_num);
      out_message (header_info);
      sprintf (header_info, "regi_size : %6lu\n", regi_size);
      out_message (header_info);
      sprintf (header_info, "transistors_num : %6lu\n", transistors_num);
      out_message (header_info);
      sprintf (header_info, "gates_num : %6lu\n", gates_num);
      out_message (header_info);
      sprintf (header_info, "reg_transfer_assign_num : %6lu\n", 
	       rt_assignement_num);
      out_message (header_info);
      sprintf (header_info, "reg_transfer_condition_num : %6lu\n",
	       rt_condition_num);
      out_message (header_info);
      sprintf (header_info, "memory_num : %6lu\n", memory_num);
      out_message (header_info);
      sprintf (header_info, "local_variable_num : %6lu\n", locvar_num);
      out_message (header_info);
      sprintf (header_info, "process_num : %6lu\n", proc_num);
      out_message (header_info);
      sprintf (header_info, "gates_array_num : %6lu\n", gates_array_num);
      out_message (header_info);
      sprintf (header_info, "delays_num : %6lu\n", delays_num);
      out_message (header_info);
      sprintf (header_info, "delay_unit : %6lu\n", delay_unit);
      out_message (header_info);
      sprintf (header_info, "min_delay : %6lu\n", min_delay);
      out_message (header_info);
      sprintf (header_info, "max_delay : %6lu\n", max_delay);
      out_message (header_info);
      sprintf (header_info, "initialization_num : %6lu\n", init_num);
      out_message (header_info);
      sprintf (header_info, "c_modules_num : %6lu\n", c_modules_num);
      out_message (header_info);
      sprintf (header_info, "c_functions_num : %6lu\n", c_functions_num);
      out_message (header_info);
      sprintf (header_info, "max_strength : %6lu\n", max_strength);
      out_message (header_info);
      sprintf (header_info, "max_size : %6lu\n", max_size);
      out_message (header_info);
      sprintf (header_info, "max_fanout : %6lu\n", max_fanout);
      out_message (header_info);
      sprintf (header_info, "max_atpg_level : %6lu\n", max_fanout);
      out_message (header_info);
      sprintf (header_info, "vdd_offset : %6ld\n", vdd_offset);
      out_message (header_info);
      sprintf (header_info, "gnd_offset : %6ld\n", gnd_offset);
      out_message (header_info);
      sprintf (header_info, "link_date  : %6lu\n", link_date);
      out_message (header_info);
      sprintf (header_info, "link_time : %6lu\n", link_time);
      out_message (header_info);
      sprintf (header_info, "creation_link_date : %6lu\n", c_link_date);
      out_message (header_info);
      sprintf (header_info, "static_size : %6lu\n", static_size);
      out_message (header_info);
      sprintf (header_info, "atpg_flag : %6lu\n", atpg_flag);
      out_message (header_info);
      sprintf (header_info, "simulation_flag : %6lu\n", sim_flag);
      out_message (header_info);
      sprintf (header_info, "gate_array_flag : %6lu\n", gate_array_flag);
      out_message (header_info);
      sprintf (header_info, "case_sensitive_flag : %6lu\n",
      out_message (header_info);
      sprintf (header_info, "total_elements_num : %6lu\n", tot_elements_num);
      out_message (header_info);
      sprintf (header_info, "total_fanin_num : %6lu\n", tot_fanin_num);
      out_message (header_info);
      sprintf (header_info, "total_fanout_num : %6lu\n", tot_fanout_num);
      out_message (header_info);
      sprintf (header_info, "total_bidirectional_num : %6lu\n", tot_bidir_num);
      out_message (header_info);
      sprintf (header_info, "nodes_fanout : %6lu\n", nodes_fo);
      out_message (header_info);
      sprintf (header_info, "nodes_fanin : %6lu\n", nodes_fi);
      out_message (header_info);
      sprintf (header_info, "nodes_connections : %6lu\n", nodes_conn);
      out_message (header_info);
      sprintf (header_info, "wires_fanout : %6lu\n", wires_fo);
      out_message (header_info);
      sprintf (header_info, "wires_fanin : %6lu\n", wires_fi);
      out_message (header_info);
      sprintf (header_info, "regi_fanout : %6lu\n", regi_fo);
      out_message (header_info);
      sprintf (header_info, "regi_fanin : %6lu\n", regi_fi);
      out_message (header_info);
      sprintf (header_info, "gates_fanout : %6lu\n", gates_fo);
      out_message (header_info);
      sprintf (header_info, "gates_fanin : %6lu\n", gates_fi);
      out_message (header_info);
      sprintf (header_info, "reg_transfer_assignment_fanout : %6lu\n", rt_a_fo);
      out_message (header_info);
      sprintf (header_info, "reg_transfer_assignment_fanin : %6lu\n", rt_a_fi);
      out_message (header_info);
      sprintf (header_info, "reg_transfer_conditional_fanout : %6lu\n", 
               rt_c_fo);
      out_message (header_info);
      sprintf (header_info, "reg_transfer_conditional_fanin : %6lu\n", rt_c_fi);
      out_message (header_info);
   }

#endif

   /*read in the delay information from the .dee file*/

   /* allocate the delay table and read the delay values */
   delay_table = (DELAY_TABLE_ELEMENT *) get_memory ((unsigned long) delays_num,
                 (unsigned long) sizeof (DELAY_TABLE_ELEMENT));

   {
      /* local variable */
      unsigned long  delay_counter;
      unsigned long  fall_tmp;
      unsigned long  rise_tmp;

      for (delay_counter = 0; delay_counter < delays_num; delay_counter++)
      {
         fscanf (net_file, "%ld %ld", &rise_tmp, &fall_tmp);
         delay_table[delay_counter].rise = rise_tmp << 1;
         delay_table[delay_counter].fall = fall_tmp << 1;
      }
   }


   /*read in the C-module section from the .dee file if it exists*/

   /* the C-module section, C-module table and C-function table are
      not yet implemeted */

   if (c_modules_num)
   {
      fclose (net_file);
      good_bye (NOT_YET_IMPLEMENTED, "C modules handling");
   }

   if (c_functions_num)
   {
      fclose (net_file);
      good_bye (NOT_YET_IMPLEMENTED, "C functions handling");
   } 


   /* processes are not implemented */
   if (proc_num)
   {
      fclose (net_file);
      good_bye (NOT_YET_IMPLEMENTED, "processes handling");
   }



   /* allocate the memory for the topological data structures.  
      1 for the element itself and 1 for the element's function */
   num_of_elements = tot_elements_num; 

   topology_vector = (TOPOLOGICAL_ELEMENT *) get_memory ((unsigned long)
                     num_of_elements, (unsigned long) sizeof 
                     (TOPOLOGICAL_ELEMENT));



   /* allocate the memory for the reference machines.  Each element 
      contains one reference machine for the function and one for each 
      output of the element. */ 
   num_of_machines = tot_elements_num + tot_output_num;   

   ref_allocation_pointer = (MACHINE *) get_memory
                             ((unsigned long) num_of_machines,
                              (unsigned long) sizeof (MACHINE)); 


   if (tot_fanin_num > 0)
      fanin_vector = (FANIN_ELEMENT *) get_memory
                             ((unsigned long) tot_fanin_num,
                              (unsigned long) sizeof (FANIN_ELEMENT));


   if (tot_fanout_num > 0)
      fanout_vector = (FANOUT_ELEMENT *) get_memory
                             ((unsigned long) tot_fanout_num,
                              (unsigned long) sizeof (FANOUT_ELEMENT));


   /* load the topological elements */

   {

      /* external functions */
      extern unsigned long  ascii_to_value();
      extern int            insert_event();


      /* local variables */
      unsigned long   te_counter;
      unsigned long   ref_machine_index;
      unsigned long   te_attribute;
      unsigned long   te_code;
      unsigned long   te_sim_level;
      unsigned long   te_atpg_level;
      unsigned long   te_dim;
      unsigned long   te_size;
      unsigned long   te_type;
      unsigned long   te_strength;
      unsigned long   te_0_strength;
      unsigned long   te_1_strength;
      unsigned long   te_delay_index;
      unsigned long   te_output_num;
      unsigned long   output_counter;
      unsigned long   te_fo_num;
      unsigned long   te_input_num;
      unsigned long   te_fi_num;
      unsigned long   te_conn_num;
      unsigned long   fo_offset;
      unsigned long   fi_offset;
      unsigned long   link_counter;
      unsigned long   link_type;
      unsigned long   link_index;
      unsigned long   fi_index;
      unsigned long   fo_index;
      unsigned long   link_b1c;
      unsigned long   link_b2c;
      unsigned long   link_b1d;
      unsigned long   link_b2d;
      unsigned long   link_item;
      char            const_string [MAX_DATA_SIZE];
      MACHINE         *ref_temp_pointer;
      MACHINE         *ref_func_temp_pointer;


      /* Initialize */
      fo_offset = fi_offset = 0;
      max_level_number = 0;
      max_eval_link_number = 0;
      te_fo_num = 0;
      ref_machine_index = 0;

    

         /* scan in general element information: 
            te_attribute  => the position of the element in the circuit
	    te_code       => what kind of element (gate, transistor, 
                             behavioral model)
  	    te_sim_level  => the simulation level
	    te_atpg_level => the automatic test pattern generation level 
         */	 

      for (te_counter = 0; te_counter < num_of_elements; te_counter++)
      {
         fscanf (net_file, "%ld %ld %ld %ld", &te_attribute, &te_code,
                                              &te_sim_level, &te_atpg_level);


         switch ( (unsigned char) te_attribute)
         {
            case MZ_INTERNAL : topology_vector[te_counter].position = INTERNAL;
                               topology_vector[te_counter].detection_point = 
							   FALSE;
                               break;

            case MZ_INPUT    : topology_vector[te_counter].position = INPUT;
                               topology_vector[te_counter].detection_point = 
							   FALSE;
                               break;

            case MZ_OUTPUT   : topology_vector[te_counter].position = OUTPUT;
                               /* the detection point will be set during the 
			           simulation */
                               topology_vector[te_counter].detection_point = 
							   FALSE;
                               break;

            case MZ_BIDIRECTIONAL : topology_vector[te_counter].position = 
								BIDIRECTIONAL;
                               topology_vector[te_counter].detection_point = 
							   FALSE;
                               break;

            case MZ_PHANTOM   : topology_vector[te_counter].position = PHANTOM;
                                topology_vector[te_counter].detection_point = 
							    FALSE;
                                break;

            default           : good_bye (BUG, "unknown element attribute");
         }


         if (max_level_number < te_sim_level)
         {
            max_level_number = te_sim_level;
         }

         topology_vector[te_counter].level = te_sim_level;
         topology_vector[te_counter].fanin_position = fanin_vector + fi_offset;
         topology_vector[te_counter].fanout_position = fanout_vector + fo_offset;
         topology_vector[te_counter].last_update_time = MAX_UNSIGNED;
         topology_vector[te_counter].stack_link = (TOPOLOGICAL_ELEMENT *) MAX_UNSIGNED;
         topology_vector[te_counter].appendix = (APPENDIX *) NULL;

#ifdef RESTORE_ENABLED
         topology_vector[te_counter].saved_list_head = (MACHINE *) NULL;
         topology_vector[te_counter].save_time = MAX_UNSIGNED;
#endif

         switch ( (unsigned char) te_code)
         {
            case MZ_NODE:
                     topology_vector[te_counter].type = NODE;

                     /* read the second record of the topological descriptor */
                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld %ld %ld",
                                       &te_dim, &te_size, &te_delay_index,
                                       &te_fo_num, &te_fi_num, &te_conn_num,
				       &te_output_num, &te_input_num);

		     if (te_dim != 1)
                     {
                        printf ("Vectorial nodes not allowed.\n");
                        good_bye (TRUE, TRUE);
                     }

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = SWITCH_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     topology_vector[te_size].driving_capabilities = te_size;

                     break;

            case MZ_WIRE:

                     /* read the second record of the topological descriptor */
                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld %ld %ld",
                                       &te_dim, &te_size, &te_delay_index,
                                       &te_fo_num, &te_fi_num, &te_conn_num,
				       &te_output_num, &te_input_num);

		     if (te_dim > MAX_DATA_SIZE)
                     {
                        printf ("Unlimited data bit-length not allowed.\n");
                        good_bye (TRUE, TRUE);
                     }

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;

                     /* is the size really meaningful? */
                     if (te_size == 0)
                     {
                        topology_vector[te_counter].type = WIRE;
                     }
                     else
                     {
                        topology_vector[te_counter].type = MEM_WIRE;
                     }

                     /* link the element to the second creation pass list
                        to store the strength of its signal read from its
                        fanin elements */
                     topology_vector[te_counter].stack_link = second_pass;
                     second_pass = &(topology_vector[te_counter]);

                     break;

            case MZ_CONST:
                     topology_vector[te_counter].type = CONSTANT;

                     /* read the second record of the topological descriptor */
                     fscanf (net_file, "%s %ld %ld %ld %ld", const_string, 
			     &te_fo_num, &te_conn_num, &te_output_num, 
			     &te_input_num);

		     te_fi_num = 0;
                     te_dim = strlen (const_string);

                     if (te_dim > MAX_DATA_SIZE)
                     {
                        printf ("Unlimited data bit-length not allowed.\n");
                        good_bye (TRUE, TRUE);
                     }

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     SET_ST_STRENGTH (topology_vector[te_counter]
                                                        .driving_capabilities,
                                      MAX_STRENGTH + 1, MAX_STRENGTH + 1);
                     break;


            case MZ_TRANSISTOR:
                     /* read the second record of the topological descriptor */
                     fscanf (net_file, "%ld %ld %ld %ld", &te_type, 
			     &te_strength, &te_output_num, &te_input_num);

                     te_fo_num = te_fi_num = te_conn_num = 0;
		      topology_vector[te_counter].data_size = 1;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     SET_ST_STRENGTH (topology_vector[te_counter]
                                                        .driving_capabilities,
                                      te_strength, te_strength);

                     switch ( (unsigned char) te_type)
                     {
                        case MZ_DEPLETION:
                               topology_vector[te_counter].type = DEPL_TRANS;
                               break;
                        case MZ_N_CHANNEL:
                               topology_vector[te_counter].type = N_CH_TRANS;
                               break;
                        case MZ_P_CHANNEL:
                               topology_vector[te_counter].type = P_CH_TRANS;
                               break;
                        default: printf ("unknown transistor type");
                                 good_bye (TRUE, TRUE);
                     }

                     break;


	    case MZ_BEHAVIORAL:
                     /* read the second record of the topological descriptor */
                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld %ld %ld",
                                      &te_type, &te_delay_index, &te_0_strength,
                                      &te_1_strength, &te_fo_num, &te_fi_num,
				      &te_output_num, &te_input_num);


		     te_conn_num = 0;
                     te_dim = 1;

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     SET_ST_STRENGTH (topology_vector[te_counter]
                                                        .driving_capabilities,
                                      te_0_strength, te_1_strength);



                     switch ( (unsigned char) te_type)
  		     {
                        case MZ_HALF_ADDER:
                                 topology_vector[te_counter].type = HALF_ADDER;
                                 break;

		        case MZ_FULL_ADDER:
                                 topology_vector[te_counter].type = FULL_ADDER;
				 break;

		        case MZ_RING_BUFFER: 
                                 topology_vector[te_counter].type = RING_BUFFER;
				 break;
		        case MZ_S27: 
                                 topology_vector[te_counter].type = S27;
				 break;
			case MZ_4TO1_MUX:
                                 topology_vector[te_counter].type = MUX_4TO1;
				 break;
			case MZ_MEMS_PSENSOR:
                                 topology_vector[te_counter].type = MEMS_PSENSOR;
				 break;
				

                        default:
                                 good_bye 
                                    (BUG, "unknown Behavioral Model");
                     }

		     break;


            case MZ_GATE:
                     /* read the second record of the topological descriptor */
                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld %ld %ld",
                                      &te_type, &te_delay_index, &te_0_strength,
                                      &te_1_strength, &te_fo_num, &te_fi_num,
				      &te_output_num, &te_input_num);

		     te_conn_num = 0;
                     te_dim = 1;

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     SET_ST_STRENGTH (topology_vector[te_counter]
                                                        .driving_capabilities,
                                      te_0_strength, te_1_strength);



                     switch ( (unsigned char) te_type)
                     {
                        case MZ_AND       :
                                 topology_vector[te_counter].type = AND;
                                 break;
                        case MZ_NAND      :
                                 topology_vector[te_counter].type = NAND;
                                 break;
                        case MZ_OR        :
                                 topology_vector[te_counter].type = OR;
                                 break;
                        case MZ_NOR       :
                                 topology_vector[te_counter].type = NOR;
                                 break;
                        case MZ_BUF       :
                                 topology_vector[te_counter].type = BUF;
                                 break;
                        case MZ_INV       :
                                 topology_vector[te_counter].type = INV;
                                 break;
                        case MZ_XOR       :
                                 topology_vector[te_counter].type = XOR;
                                 break;
                        case MZ_XNOR      :
                                 topology_vector[te_counter].type = XNOR;
                                 break;
                        case MZ_TGP       :
                                 topology_vector[te_counter].type = TGP      ;
                                 break;
                        case MZ_TGN       :
                                 topology_vector[te_counter].type = TGN      ;
                                 break;
                        case MZ_MTGP      :
                                 topology_vector[te_counter].type = MTGP     ;
                                 break;
                        case MZ_MTGN      :
                                 topology_vector[te_counter].type = MTGN     ;
                                 break;
                        case MZ_BUS_TRI   :
                                 topology_vector[te_counter].type = BUS_TRI  ;
                                 break;
                        case MZ_BUS_MOS   :
                                 topology_vector[te_counter].type = BUS_MOS  ;
                                 break;
                        case MZ_BUS_UP    :
                                 topology_vector[te_counter].type = BUS_UP   ;
                                 break;
                        case MZ_BUS_DOWN  :
                                 topology_vector[te_counter].type = BUS_DOWN ;
                                 break;
                        case MZ_AND_TRI   :
                                 topology_vector[te_counter].type = AND_TRI  ;
                                 break;
                        case MZ_NAND_TRI  :
                                 topology_vector[te_counter].type = NAND_TRI ;
                                 break;
                        case MZ_OR_TRI    :
                                 topology_vector[te_counter].type = OR_TRI   ;
                                 break;
                        case MZ_NOR_TRI   :
                                 topology_vector[te_counter].type = NOR_TRI  ;
                                 break;
                        case MZ_BUF_TRI   :
                                 topology_vector[te_counter].type = BUF_TRI  ;
                                 break;
                        case MZ_INV_TRI   :
                                 topology_vector[te_counter].type = INV_TRI  ;
                                 break;
                        case MZ_AND_TRIN  :
                                 topology_vector[te_counter].type = AND_TRIN ;
                                 break;
                        case MZ_NAND_TRIN :
                                 topology_vector[te_counter].type = NAND_TRIN;
                                 break;
                        case MZ_OR_TRIN   :
                                 topology_vector[te_counter].type = OR_TRIN  ;
                                 break;
                        case MZ_NOR_TRIN  :
                                 topology_vector[te_counter].type = NOR_TRIN ;
                                 break;
                        case MZ_BUF_TRIN  :
                                 topology_vector[te_counter].type = BUF_TRIN ;
                                 break;
                        case MZ_INV_TRIN  :
                                 topology_vector[te_counter].type = INV_TRIN ;
                                 break;
                        case MZ_AND_OC    :
                                 topology_vector[te_counter].type = AND_OC   ;
                                 break;
                        case MZ_NAND_OC   :
                                 topology_vector[te_counter].type = NAND_OC  ;
                                 break;
                        case MZ_OR_OC     :
                                 topology_vector[te_counter].type = OR_OC    ;
                                 break;
                        case MZ_NOR_OC    :
                                 topology_vector[te_counter].type = NOR_OC   ;
                                 break;
                        case MZ_BUF_OC    :
                                 topology_vector[te_counter].type = BUF_OC   ;
                                 break;
                        case MZ_INV_OC    :
                                 topology_vector[te_counter].type = INV_OC   ;
                                 break;
                        case MZ_AND_OE    :
                                 topology_vector[te_counter].type = AND_OE   ;
                                 break;
                        case MZ_NAND_OE   :
                                 topology_vector[te_counter].type = NAND_OE  ;
                                 break;
                        case MZ_OR_OE     :
                                 topology_vector[te_counter].type = OR_OE    ;
                                 break;
                        case MZ_NOR_OE    :
                                 topology_vector[te_counter].type = NOR_OE   ;
                                 break;
                        case MZ_BUF_OE    :
                                 topology_vector[te_counter].type = BUF_OE   ;
                                 break;
                        case MZ_INV_OE    :
                                 topology_vector[te_counter].type = INV_OE   ;
                                 break;
                        case MZ_SRNOR     :
                                 topology_vector[te_counter].type = SRNOR    ;
                                 break;
                        case MZ_SRNAND    :
                                 topology_vector[te_counter].type = SRNAND   ;
                                 break;
                        case MZ_MUX       :
                                 topology_vector[te_counter].type = MUX      ;
                                 break;


                        /* flip-flops */

                        case MZ_DFF_PET   :
                                 topology_vector[te_counter].type = DFF_PET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_PET   :
                                 topology_vector[te_counter].type = TFF_PET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_PET_C :
                                 topology_vector[te_counter].type = DFF_PET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_PET_C :
                                 topology_vector[te_counter].type = TFF_PET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_PET_S :
                                 topology_vector[te_counter].type = DFF_PET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_PET_S :
                                 topology_vector[te_counter].type = TFF_PET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_PET_SC:
                                 topology_vector[te_counter].type = DFF_PET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_PET_SC :
                                 topology_vector[te_counter].type = TFF_PET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_NET   :
                                 topology_vector[te_counter].type = DFF_NET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_NET   :
                                 topology_vector[te_counter].type = TFF_NET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_NET_C :
                                 topology_vector[te_counter].type = DFF_NET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_NET_C :
                                 topology_vector[te_counter].type = TFF_NET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_NET_S :
                                 topology_vector[te_counter].type = DFF_NET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_NET_S :
                                 topology_vector[te_counter].type = TFF_NET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_NET_SC:
                                 topology_vector[te_counter].type = DFF_NET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_NET_SC :
                                 topology_vector[te_counter].type = TFF_NET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_PET    :
                                 topology_vector[te_counter].type = JKFF_PET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_PET_C  :
                                 topology_vector[te_counter].type = JKFF_PET_C;


                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_PET_S  :
                                 topology_vector[te_counter].type = JKFF_PET_S;


                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_PET_SC  :
                                 topology_vector[te_counter].type = JKFF_PET_SC;


                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 4].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_NET    :
                                 topology_vector[te_counter].type = JKFF_NET;


                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_NET_C  :
                                 topology_vector[te_counter].type = JKFF_NET_C;


                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_NET_S  :
                                 topology_vector[te_counter].type = JKFF_NET_S;


                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_NET_SC  :
                                 topology_vector[te_counter].type = JKFF_NET_SC;


                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 4].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_LATCH_HIGH  :
                                 topology_vector[te_counter].type = LATCH_HIGH;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_LATCH_LOW  :
                                 topology_vector[te_counter].type = LATCH_LOW;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        default           :
                                 good_bye 
                                    (BUG, "unknown gate or register type");
                     }

                     break;

            default: good_bye (NOT_YET_IMPLEMENTED,
                             "the management of some elements in the network");

         }  /* end of switch (te_code) */





#ifdef VECTORIAL
         /* for debug purpose, the data size of each element may be 
	    to MAX_DATA_SIZE */
         topology_vector[te_counter].data_size = MAX_DATA_SIZE;
#endif


         /* read and store the links */
         if (max_eval_link_number < te_fi_num)
         {
            max_eval_link_number = te_fi_num;
         }

         /* fanout list */
         for (link_counter = 0; link_counter < te_fo_num; link_counter++)
         {
            fscanf (net_file, "%ld %ld", &link_type, &link_index);

            /* NOTE: some fanout links, e.g., the non triggering ones, 
                     should be hidden, to avoid the propagation of the 
                     activity on them. However, not enough information are 
                     generated by the linker to do this. */

            fanout_vector[fo_offset].link = topology_vector + link_index;
            fanout_vector[fo_offset].attribute = VOID;

            if (te_code == MZ_RT_ASSIGNEMENT)
            {
               /* for now, ignore the information */
               fscanf (net_file, "%ld %ld %ld %ld", &link_b1c, &link_b2c,
                                                    &link_b1d, &link_b2d);
	       fanout_vector[fo_offset].fo_index = 0;
            }
            else
            {
               /*use fo_index to represent the index of the fanout.   
		 Conceptually: think of the fanout as an array of outputs */

               fscanf (net_file, "%ld %ld", &fo_index, &link_item);
	       fanout_vector[fo_offset].fo_index = fo_index;
            }
            fo_offset++;
	  }


         /* fanin list */

         /* input order for flip-flops in MOZART is really ugly. Now it 
            is time to change. */
         
         switch (topology_vector[te_counter].type)
         {
            case DFF_NET_C:
            case DFF_NET_S:
            case DFF_PET_C:
            case DFF_PET_S:
            case TFF_NET_C:
            case TFF_NET_S:
            case TFF_PET_C:
            case TFF_PET_S:
                  /* preset or clear: 3rd link */
                  fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                       &fi_index, &link_item);
      
                  if (fi_index != 0)
                  {
                     good_bye (NOT_YET_IMPLEMENTED, "bit extraction");
                  }
                  fanin_vector[fi_offset+2].link = 
                                          topology_vector + link_index;

                  /* data: 1st link */
                  fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                       &fi_index, &link_item);
      
                  if (fi_index != 0)
                  {
                     good_bye (NOT_YET_IMPLEMENTED, "bit extraction");
                  }
                  fanin_vector[fi_offset].link = 
                                          topology_vector + link_index;

                  /* clock: 2nd link */
                  fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                       &fi_index, &link_item);
      
                  if (fi_index != 0)
                  {
                     good_bye (NOT_YET_IMPLEMENTED, "bit extraction");
		   }
                  fanin_vector[fi_offset+1].link = 
                                          topology_vector + link_index;

                  fi_offset += 3;
                  break;

            case DFF_PET_SC:
            case DFF_NET_SC:
            case TFF_PET_SC:
            case TFF_NET_SC:
                  /* preset: 3rd link */
                  fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                       &fi_index, &link_item);
      
                  if (fi_index != 0)
                  {
                     good_bye (NOT_YET_IMPLEMENTED, "bit extraction");
                  }
                  fanin_vector[fi_offset+2].link = 
                                          topology_vector + link_index;
		  fanin_vector[fi_offset+2].fi_index = 0;

                  /* clear: 4th link */
                  fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                       &fi_index, &link_item);
      
                  if (fi_index != 0)
                  {
                     good_bye (NOT_YET_IMPLEMENTED, "bit extraction");
                  }
                  fanin_vector[fi_offset+3].link = 
                                          topology_vector + link_index;
 		  fanin_vector[fi_offset+3].fi_index = 0;

                  /* data: 1st link */
                  fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                       &fi_index, &link_item);
      
                  if (fi_index != 0)
                  {
                     good_bye (NOT_YET_IMPLEMENTED, "bit extraction");
                  }
                  fanin_vector[fi_offset].link = 
                                          topology_vector + link_index;
		  fanin_vector[fi_offset].fi_index = 0;


                  /* clock: 2nd link */
                  fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                       &fi_index, &link_item);
      
                  if (fi_index != 0)
                  {
                     good_bye (NOT_YET_IMPLEMENTED, "bit extraction");
		   }
                  fanin_vector[fi_offset+1].link = 
                                          topology_vector + link_index;

		  fanin_vector[fi_offset+1].fi_index = 0;

                  fi_offset += 4;
                  break;

            case DFF_PET:
            case DFF_NET:
            default:
               for (link_counter = 0; link_counter < te_fi_num; link_counter++)
               {
                  fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                       &fi_index, &link_item);
      
                  fanin_vector[fi_offset].link = topology_vector + link_index;

		  /* defines which fanin is it out of the inputs for the 
		     element. */
		  fanin_vector[fi_offset].fi_index = fi_index;
                  fi_offset++;
		}
               break;
		}


         /* the computational appendix for now cannot be loaded */

         if ((te_code == MZ_RT_ASSIGNEMENT) || (te_code == MZ_RT_CONDITION))
         {
            good_bye (NOT_YET_IMPLEMENTED,
                            "RT assignements and conditions handling");
	  }


         /*update this index upon looping*/
 	 ref_machine_index = ref_machine_index + te_output_num + 1;

      }  /* end of one element loading */

   }  /* end of all the elements loading */

   
   /* start the second pass to store the strength in the wire descriptors */

   while (second_pass != NULL)
   {
      unsigned long        link_cnt;
      FANIN_ELEMENT        *fi_link;
      unsigned             driving_capabilities;
      TOPOLOGICAL_ELEMENT  *next_second_pass;

      /* beginning of executable part */

      /* read the strength of all the elements writing on the wire */

      if (second_pass->fanin_number == 0)
      {
         driving_capabilities = 0;
      }
      else
      {
         fi_link = second_pass->fanin_position;
         driving_capabilities = fi_link->link->driving_capabilities;
      }

      for (link_cnt = 1; link_cnt < second_pass->fanin_number; link_cnt++)
      {
         if (driving_capabilities != fi_link->link->driving_capabilities)
         {
            printf("Different strength on a single wire. Max is forced.\n");
            SET_ST_STRENGTH (driving_capabilities, 
                             MAX_STRENGTH + 1, MAX_STRENGTH + 1);
            break;
         }
         fi_link++;

      }  /* end of loop on fanin */

      second_pass->driving_capabilities = driving_capabilities;

      next_second_pass = second_pass->stack_link;
      second_pass->stack_link = (TOPOLOGICAL_ELEMENT *) MAX_UNSIGNED;
      second_pass = next_second_pass;

   }


   /* initialize VDD and GND if present */

   {

      /* beginning of executable part */

      if (vdd_offset != MAX_UNSIGNED)
      {
         /* modify the element position code */
         topology_vector[vdd_offset].position = VDD;
      }

      if (gnd_offset != MAX_UNSIGNED)
      {
         /* modify the element position code */
         topology_vector[gnd_offset].position = GND;
      }
   }



   /* load the initialization section.
      Note that since the initialization events are inserted in the
      time queue, it must be initialized before this routine is run. */

   {

      /* local variables */

      unsigned long  init_values_counter;
      unsigned long  init_element_counter;
      unsigned char  init_string [MAX_DATA_SIZE];
      unsigned long  init_internal_value;
      unsigned long  init_index;


      /* beginning of executable part */

      for (init_values_counter = 0; init_values_counter < init_num;
           init_values_counter++)
      {
         /* read the initialization value and the number of elements
            to be initialized */

         fscanf (net_file, "%s %ld", init_string, &init_element_counter);

         for ( ; init_element_counter > 0; init_element_counter--)
         {
            /* read the index of the element to be initialized. */

            fscanf (net_file, "%ld", &init_index);

            /* ignore the information */

         }
      }
   }



   return;

}  /* end of load_net */
